#!/usr/bin/bash
#
# Deletes the files that are generated by running test and build commands for
# this repository.

set -e -x

find . \
    -name .tox -o -name build -o -name dist \
    -o -name _autosummary -o -name __pycache__ \
    -o -name .coverage -o -name '.coverage.*' -o -name htmlcov \
    | xargs -r rm -rf
